import psutil
from bsb import Simulation, config, types

from .cell import ArborCell
from .connection import ArborConnection
from .device import ArborDevice


@config.node
class ArborSimulation(Simulation):
    resolution = config.attr(type=types.float(min=0.0), default=0.1)
    profiling = config.attr(type=bool)
    cell_models = config.dict(type=ArborCell, required=True)
    connection_models = config.dict(type=ArborConnection, required=True)
    devices = config.dict(type=ArborDevice, required=True)

    @config.property(default=1)
    def threads(self):
        return self._threads

    @threads.setter
    def threads(self, value):
        self._threads = value if value != "all" else psutil.cpu_count(logical=False)



import abc
import typing

import arbor
from bsb import CellModel, ConfigurationError, config, types

from .adapter import SingleReceiverCollection

if typing.TYPE_CHECKING:
    from bsb import PlacementSet


@config.dynamic(
    attr_name="model_strategy",
    auto_classmap=True,
    required=True,
    classmap_entry=None,
)
class ArborCell(CellModel):
    gap = config.attr(type=bool, default=False)
    model = config.attr(type=types.class_(), required=True)

    @abc.abstractmethod
    def cache_population_data(self, simdata, ps: "PlacementSet"):
        pass

    @abc.abstractmethod
    def discard_population_data(self):
        pass

    @abc.abstractmethod
    def get_prefixed_catalogue(self):
        pass

    @abc.abstractmethod
    def get_cell_kind(self, gid):
        pass

    @abc.abstractmethod
    def make_receiver_collection(self):
        pass

    def get_description(self, gid):
        morphology, labels, decor = self.model.cable_cell_template()
        labels = self._add_labels(gid, labels, morphology)
        decor = self._add_decor(gid, decor)
        cc = arbor.cable_cell(morphology, labels, decor)
        return cc


@config.node
class LIFCell(ArborCell, classmap_entry="lif"):
    model = config.unset()
    constants = config.dict(type=types.any_())

    def cache_population_data(self, simdata, ps: "PlacementSet"):
        pass

    def discard_population_data(self):
        pass

    def get_prefixed_catalogue(self):
        return None, None

    def get_cell_kind(self, gid):
        return arbor.cell_kind.lif

    def get_description(self, gid):
        cell = arbor.lif_cell(f"-1_-1", f"-1_-1_0")
        try:
            for k, v in self.constants.items():
                setattr(cell, k, v)
        except AttributeError:
            node_name = type(self).constants.get_node_name(self)
            raise ConfigurationError(
                f"'{k}' is not a valid LIF parameter in '{node_name}'."
            ) from None
        return cell

    def make_receiver_collection(self):
        return SingleReceiverCollection()



import arbor
from bsb import config

from ..connection import Receiver
from ..device import ArborDevice


@config.node
class PoissonGenerator(ArborDevice, classmap_entry="poisson_generator"):
    record = config.attr(type=bool, default=True)
    rate = config.attr(type=float, required=True)
    weight = config.attr(type=float, required=True)
    delay = config.attr(type=float, required=True)

    def implement_probes(self, simdata, gid):
        return []

    def implement_generators(self, simdata, gid):
        target = Receiver(self, None, [-1, -1], [-1, -1], 0).on()
        gen = arbor.event_generator(
            target, self.weight, arbor.poisson_schedule(0, self.rate / 1000, gid)
        )
        return [gen]



"""
Arbor simulation adapter for the BSB framework
"""

from bsb import SimulationBackendPlugin

from . import devices
from .adapter import ArborAdapter
from .simulation import ArborSimulation

__version__ = "0.0.0-b3"
__plugin__ = SimulationBackendPlugin(Simulation=ArborSimulation, Adapter=ArborAdapter)



import abc

import arbor
from bsb import DeviceModel, Targetting, config, types


@config.dynamic(attr_name="device", auto_classmap=True, classmap_entry=None)
class ArborDevice(DeviceModel):
    targetting = config.attr(type=Targetting, required=True)
    resolution = config.attr(type=float)
    sampling_policy = config.attr(type=types.in_(["exact"]))

    def __init__(self, **kwargs):
        self._probe_ids = []

    def __boot__(self):
        self.resolution = self.resolution or self.simulation.resolution

    def register_probe_id(self, gid, tag):
        self._probe_ids.append((gid, tag))

    def prepare_samples(self, simdata):
        self._handles = [
            self.sample(simdata.arbor_sim, probe_id) for probe_id in self._probe_ids
        ]

    def sample(self, sim, probe_id):
        schedule = arbor.regular_schedule(self.resolution)
        sampling_policy = getattr(arbor.sampling_policy, self.sampling_policy)
        return sim.sample(probe_id, schedule, sampling_policy)

    def get_samples(self, sim):
        return [sim.samples(handle) for handle in self._handles]

    def get_meta(self):
        attrs = ("name", "sampling_policy", "resolution")
        return dict(zip(attrs, (getattr(self, attr) for attr in attrs)))

    @abc.abstractmethod
    def implement_probes(self, simdata, target):
        pass

    @abc.abstractmethod
    def implement_generators(self, simdata, target):
        pass
